import re
from functools import lru_cache

import requests

from grayskull.cli import CLIConfig


def build_package_url(channel_or_url: str, pkg_name: str) -> str:
    """Build the URL to check for package availability.

    :param channel_or_url: Channel name or full URL
    :param pkg_name: Package name
    :return: Full URL to check for package availability
    """
    # Check if the channel is a full URL
    if re.match(r"^https?://", channel_or_url):
        return f"{channel_or_url.rstrip('/')}/{pkg_name}/files"
    else:
        # Default to anaconda.org if not a full URL
        return f"https://anaconda.org/{channel_or_url}/{pkg_name}/files"


@lru_cache(maxsize=35)
def is_pkg_available(pkg_name: str, channel: str = None) -> bool:
    """Verify if the package is available on Anaconda or custom package indexes.

    :param pkg_name: Package name
    :param channel: Anaconda channel
        or full URL (if None, will use channels from CLIConfig)
    :return: Return True if the package is present on any of the given channels or URLs
    """
    channels_to_check = [channel] if channel else CLIConfig().package_indexes

    for channel_to_check in channels_to_check:
        try:
            url = build_package_url(channel_to_check, pkg_name)
            response = requests.get(
                url=url,
                allow_redirects=False,
            )
            if response.status_code == 200:
                return True
        except requests.exceptions.RequestException:
            continue
    return False


def normalize_pkg_name(pkg_name: str) -> str:
    if is_pkg_available(pkg_name):
        return pkg_name
    if is_pkg_available(pkg_name.replace("-", "_")):
        return pkg_name.replace("-", "_")
    elif is_pkg_available(pkg_name.replace("_", "-")):
        return pkg_name.replace("_", "-")
    return pkg_name



import sys
from collections import defaultdict
from collections.abc import Iterator
from functools import singledispatch
from pathlib import Path

from grayskull.strategy.parse_poetry_version import (
    combine_conda_selectors,
    encode_poetry_platform_to_selector_item,
    encode_poetry_python_version_to_selector_item,
    encode_poetry_version,
)
from grayskull.utils import nested_dict

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


class InvalidPoetryDependency(BaseException):
    pass


@singledispatch
def get_constrained_dep(dep_spec: list | str | dict, dep_name: str) -> str:
    raise InvalidPoetryDependency(
        "Expected Poetry dependency specification to be of type list, str or dict, "
        f"received {type(dep_spec).__name__}"
    )


@get_constrained_dep.register
def __get_constrained_dep_dict(
    dep_spec: dict, dep_name: str
) -> Iterator[str, None, None]:
    """
    Yield a dependency entry in conda format from a Poetry entry
    with version, python version, and platform

    Example:
        dep_spec:
            {"version": "^1.5", "python": ">=3.8,<3.12", "platform": "darwin"},
        dep_name:
            "pandas",
        result yield:
            "pandas >=1.5.0,<2.0.0  # [py>=38 and py<312 and osx]"
    """
    conda_version = encode_poetry_version(dep_spec.get("version", ""))
    if conda_version:
        conda_version = f" {conda_version}"
    python_selector = encode_poetry_python_version_to_selector_item(
        dep_spec.get("python", "")
    )
    platform_selector = encode_poetry_platform_to_selector_item(
        dep_spec.get("platform", "")
    )
    conda_selector = combine_conda_selectors(python_selector, platform_selector)
    yield f"{dep_name}{conda_version}{conda_selector}".strip()


@get_constrained_dep.register
def __get_constrained_dep_str(
    dep_spec: str, dep_name: str
) -> Iterator[str, None, None]:
    conda_version = encode_poetry_version(dep_spec)
    yield f"{dep_name} {conda_version}"


@get_constrained_dep.register
def __get_constrained_dep_list(
    dep_spec_list: list, dep_name: str
) -> Iterator[str, None, None]:
    for dep_spec in dep_spec_list:
        yield from get_constrained_dep(dep_spec, dep_name)


def encode_poetry_deps(poetry_deps: dict) -> tuple[list, list]:
    run = []
    run_constrained = []
    for dep_name, dep_spec in poetry_deps.items():
        for constrained_dep in get_constrained_dep(dep_spec, dep_name):
            try:
                assert dep_spec.get("optional", False)
                run_constrained.append(constrained_dep)
            except (AttributeError, AssertionError):
                run.append(constrained_dep)
    return run, run_constrained


def add_poetry_metadata(metadata: dict, toml_metadata: dict) -> dict:
    if not is_poetry_present(toml_metadata):
        return metadata

    poetry_metadata = toml_metadata["tool"]["poetry"]
    poetry_deps = poetry_metadata.get("dependencies", {})
    req_run, req_run_constrained = encode_poetry_deps(poetry_deps)

    # add dependencies
    metadata["requirements"].setdefault("run", [])
    metadata["requirements"]["run"].extend(req_run)

    # add optional dependencies
    if len(req_run_constrained):
        metadata["requirements"].setdefault("run_constrained", [])
        metadata["requirements"]["run_constrained"].extend(req_run_constrained)

    host_metadata = metadata["requirements"].get("host", [])
    if toml_metadata["tool"].get("poetry", {}).get("scripts"):
        metadata["build"]["entry_points"] = []
        for entry_name, entry_path in toml_metadata["tool"]["poetry"][
            "scripts"
        ].items():
            metadata["build"]["entry_points"].append(f"{entry_name} = {entry_path}")
    if "poetry" not in host_metadata and "poetry-core" not in host_metadata:
        metadata["requirements"]["host"] = host_metadata + ["poetry-core"]

    poetry_test_deps = (
        poetry_metadata.get("group", {}).get("test", {}).get("dependencies", {})
    )
    # add required test dependencies and ignore optional test dependencies, as
    # there doesn't appear to be a way to specify them in Conda recipe metadata.
    test_reqs, _ = encode_poetry_deps(poetry_test_deps)
    if test_reqs:
        extra_metadata = metadata["requirements"].setdefault("extra", {})
        for name in ("test", "tests", "testing"):
            if name in extra_metadata:
                extra_metadata[name] += test_reqs
                break
        else:
            extra_metadata["test"] = test_reqs
    return metadata


def is_poetry_present(toml_metadata: dict) -> bool:
    return "poetry" in toml_metadata.get("tool", {})


def is_flit_present(toml_metadata: dict) -> bool:
    return "flit" in toml_metadata.get("tool", {})


def add_flit_metadata(metadata: dict, toml_metadata: dict) -> dict:
    if not is_flit_present(toml_metadata):
        return metadata

    flit_metadata = toml_metadata["tool"]["flit"]
    flit_scripts = flit_metadata.get("scripts", {})
    for entry_name, entry_path in flit_scripts.items():
        if "build" not in metadata:
            metadata["build"] = {}
        if "entry_points" not in metadata["build"]:
            metadata["build"]["entry_points"] = []
        metadata["build"]["entry_points"].append(f"{entry_name} = {entry_path}")
    return metadata


def is_pep725_present(toml_metadata: dict):
    return "external" in toml_metadata


def get_pep725_mapping(purl: str):
    """This function maps a PURL to the name in the conda ecosystem. It is expected
    that this will be provided on a per-ecosystem basis (such as by conda-forge)"""

    package_mapping = {
        "virtual:compiler/c": "{{ compiler('c') }}",
        "virtual:compiler/cpp": "{{ compiler('cxx') }}",
        "virtual:compiler/fortran": "{{ compiler('fortran') }}",
        "virtual:compiler/rust": "{{ compiler('rust') }}",
        "virtual:interface/blas": "{{ blas }}",
        "pkg:generic/boost": "boost-cpp",
        "pkg:generic/brial": "brial",
        "pkg:generic/cddlib": "cddlib",
        "pkg:generic/cliquer": "cliquer",
        "pkg:generic/ecl": "ecl",
        "pkg:generic/eclib": "eclib",
        "pkg:generic/ecm": "ecm",
        "pkg:generic/fflas-ffpack": "fflas-ffpack",
        "pkg:generic/fplll": "fplll",
        "pkg:generic/flint": "libflint",
        "pkg:generic/libgd": "libgd",
        "pkg:generic/gap": "gap-defaults",
        "pkg:generic/gfan": "gfan",
        "pkg:generic/gmp": "gmp",
        "pkg:generic/giac": "giac",
        "pkg:generic/givaro": "givaro",
        "pkg:generic/pkg-config": "pkg-config",
        "pkg:generic/glpk": "glpk",
        "pkg:generic/gsl": "gsl",
        "pkg:generic/iml": "iml",
        "pkg:generic/lcalc": "lcalc",
        "pkg:generic/libbraiding": "libbraiding",
        "pkg:generic/libhomfly": "libhomfly",
        "pkg:generic/lrcalc": "lrcalc",
        "pkg:generic/libpng": "libpng",
        "pkg:generic/linbox": "linbox",
        "pkg:generic/m4ri": "m4ri",
        "pkg:generic/m4rie": "m4rie",
        "pkg:generic/mpc": "mpc",
        "pkg:generic/mpfi": "mpfi",
        "pkg:generic/mpfr": "mpfr",
        "pkg:generic/maxima": "maxima",
        "pkg:generic/nauty": "nauty",
        "pkg:generic/ntl": "ntl",
        "pkg:generic/pari": "pari",
        "pkg:generic/pari-elldata": "pari-elldata",
        "pkg:generic/pari-galdata": "pari-galdata",
        "pkg:generic/pari-seadata": "pari-seadata",
        "pkg:generic/palp": "palp",
        "pkg:generic/planarity": "planarity",
        "pkg:generic/ppl": "ppl",
        "pkg:generic/primesieve": "primesieve",
        "pkg:generic/primecount": "primecount",
        "pkg:generic/qhull": "qhull",
        "pkg:generic/rw": "rw",
        "pkg:generic/singular": "singular",
        "pkg:generic/symmetrica": "symmetrica",
        "pkg:generic/sympow": "sympow",
        "pkg:generic/lrslib": "lrslib",
        "pkg:generic/bliss": "bliss",
        "pkg:generic/coxeter3": "# coxeter3 # Not available on conda-forge",
        "pkg:generic/mcqd": "# mcqd # Not available on conda-forge",
        "pkg:generic/meataxe": "# meataxe # Not available on conda-forge",
        "pkg:generic/sirocco": "sirocco",
        "pkg:generic/tdlib": "# tdlib # Not available on conda-forge",
        "pkg:generic/4ti2": "4ti2",
        "pkg:generic/benzene": "# benzene # Not available on conda-forge",
        "pkg:generic/buckygen": "# buckygen # Not available on conda-forge",
        "pkg:generic/csdp": "# csdp # Not available on conda-forge",
        "pkg:generic/frobby": "# frobby # Not available on conda-forge",
        "pkg:generic/kenzo": "# kenzo # Not available on conda-forge",
        "pkg:generic/latte-integrale": "latte-integrale",
        "pkg:generic/plantri": "# plantri # Not available on conda-forge",
        "pkg:generic/qepcad": "# qepcad # Not available on conda-forge",
        "pkg:generic/tides": "# tides # Not available on conda-forge",
        "pkg:generic/tachyon": "tachyon",
        "pkg:generic/sagemath-elliptic-curves": "sagemath-db-elliptic-curves",
        "pkg:generic/sagemath-polytopes-db": "sagemath-db-polytopes",
        "pkg:generic/sagemath-graphs": "sagemath-db-graphs",
    }
    return package_mapping.get(purl, purl)


def add_pep725_metadata(metadata: dict, toml_metadata: dict):
    if not is_pep725_present(toml_metadata):
        return metadata

    externals = toml_metadata["external"]
    # each of these is a list of PURLs. For each one we find,
    # we need to map it to the the conda ecosystem
    requirements = metadata.get("requirements", {})
    section_map = (
        ("build", "build-requires"),
        ("host", "host-requires"),
        ("run", "dependencies"),
    )
    for conda_section, pep725_section in section_map:
        requirements.setdefault(conda_section, [])
        requirements[conda_section].extend(
            [get_pep725_mapping(purl) for purl in externals.get(pep725_section, [])]
        )
        optional_features = externals.get(f"optional-{pep725_section}", {})
        for feature_name, feature_deps in optional_features.items():
            requirements[conda_section].append(
                f'# OPTIONAL dependencies from feature "{feature_name}"'
            )
            requirements[conda_section].extend(
                [get_pep725_mapping(purl) for purl in feature_deps]
            )
        if not requirements[conda_section]:
            del requirements[conda_section]

    if requirements:
        metadata["requirements"] = requirements
    return metadata


def get_all_toml_info(path_toml: Path | str) -> dict:
    with open(path_toml, "rb") as f:
        toml_metadata = tomllib.load(f)
    toml_metadata = defaultdict(dict, toml_metadata)
    metadata = nested_dict()
    toml_project = toml_metadata.get("project", {}) or {}
    metadata["requirements"]["host"] = toml_metadata["build-system"].get("requires", [])
    metadata["requirements"]["run"] = toml_project.get("dependencies", [])
    metadata["requirements"]["extra"] = toml_project.get("optional-dependencies", {})
    license = toml_project.get("license")
    if isinstance(license, dict):
        license = license.get("text", "")
    metadata["about"]["license"] = license

    tom_urls = toml_project.get("urls", {})
    if homepage := tom_urls.get("Homepage"):
        metadata["about"]["home"] = homepage
    if dev_url := tom_urls.get("Source"):
        metadata["about"]["dev_url"] = dev_url

    if toml_project.get("requires-python"):
        py_constrain = f"python {toml_project['requires-python']}"
        metadata["requirements"]["host"].append(py_constrain)
        metadata["requirements"]["run"].append(py_constrain)

    if toml_project.get("scripts"):
        metadata["build"]["entry_points"] = []
        for entry_name, entry_path in toml_project.get("scripts", {}).items():
            metadata["build"]["entry_points"].append(f"{entry_name} = {entry_path}")
    if all_urls := toml_project.get("urls"):
        metadata["about"]["dev_url"] = all_urls.get("Source", None)
        metadata["about"]["home"] = all_urls.get("Homepage", None)
    metadata["about"]["summary"] = toml_project.get("description")
    metadata["name"] = metadata.get("name") or toml_project.get("name")

    add_poetry_metadata(metadata, toml_metadata)
    add_flit_metadata(metadata, toml_metadata)
    add_pep725_metadata(metadata, toml_metadata)

    return metadata



try:
    from typing import Self
except ImportError:
    from typing import TypeVar

    Self = TypeVar("Self", bound="CLIConfig")

import progressbar

WIDGET_BAR_DOWNLOAD = [
    progressbar.Percentage(),
    " ",
    progressbar.ETA(format="%(eta)8s"),
    " ",
    progressbar.AdaptiveTransferSpeed(),
    progressbar.Bar(),
]


class CLIConfig:
    __instance: Self | None = None

    def __new__(
        cls,
        stdout: bool = False,
        list_missing_deps: bool = False,
        package_indexes: list[str] = None,
    ):
        if CLIConfig.__instance is None:
            CLIConfig.__instance = object.__new__(cls)
            CLIConfig.__instance.stdout = stdout
            CLIConfig.__instance.list_missing_deps = list_missing_deps
            CLIConfig.__instance.package_indexes = package_indexes or ["conda-forge"]
        return CLIConfig.__instance



import re

from packaging.version import Version

VERSION_REGEX = re.compile(
    r"""^[vV]?
        (?P<major>0|[1-9]\d*)
        (\.
        (?P<minor>0|[1-9]\d*)
        (\.
            (?P<patch>0|[1-9]\d*)
        )?
        )?$
    """,
    re.VERBOSE,
)


class InvalidVersion(BaseException):
    pass


def parse_version(version: str) -> dict[str, int | None]:
    """
    Parses a version string (not necessarily semver) to a dictionary with keys
    "major", "minor", and "patch". "minor" and "patch" are possibly None.

    >>> parse_version("0")
    {'major': 0, 'minor': None, 'patch': None}
    >>> parse_version("1")
    {'major': 1, 'minor': None, 'patch': None}
    >>> parse_version("1.2")
    {'major': 1, 'minor': 2, 'patch': None}
    >>> parse_version("1.2.3")
    {'major': 1, 'minor': 2, 'patch': 3}
    """
    match = VERSION_REGEX.search(version)
    if not match:
        raise InvalidVersion(f"Could not parse version {version}.")

    return {
        key: None if value is None else int(value)
        for key, value in match.groupdict().items()
    }


def get_padded_base_version(version: str | Version) -> str:
    """
    Returns the same PEP440 version padded with zeroes if
    minor or micro are not specified.

    >>> get_padded_base_version("0.2.3")
    '0.2.3'
    >>> get_padded_base_version("1")
    '1.0.0'
    >>> get_padded_base_version("1.2")
    '1.2.0'
    >>> get_padded_base_version("1.2.3")
    '1.2.3'
    >>> get_padded_base_version("1.2.3.post1")
    '1.2.3.post1'
    >>> get_padded_base_version("2!1.2.post1")
    '2!1.2.0.post1'
    """
    if not isinstance(version, Version):
        version = Version(version)

    # Start with the normalized release
    floor = f"{version.major}.{version.minor}.{version.micro}"

    # Add other components as they appear
    if version.epoch is not None and version.epoch > 0:
        floor = f"{version.epoch}!{floor}"  # Add epoch if present
    if version.pre is not None:
        floor += (
            f"{version.pre[0]}{version.pre[1]}"  # Add pre-release (e.g., a1, b1, rc1)
        )
    if version.post is not None:
        floor += f".post{version.post}"  # Add post-release (e.g., .post1)
    if version.dev is not None:
        floor += f".dev{version.dev}"  # Add development release (e.g., .dev1)
    if version.local is not None:
        floor += f"+{version.local}"  # Add local metadata (e.g., +local)
    return floor


def get_caret_ceiling(version: str | Version) -> str:
    """
    Accepts a Poetry caret version and returns the exclusive version ceiling.

    Targets that are invalid semver strings (e.g. "1.2", "0") are handled
    according to the Poetry caret requirements specification, which is based on
    whether the major version is 0:

    - If the major version is 0, the ceiling is determined by bumping the
    rightmost specified digit and then coercing it to semver.
    Example: 0 => 1.0.0, 0.1 => 0.2.0, 0.1.2 => 0.1.3

    - If the major version is not 0, the ceiling is determined by
    coercing it to semver and then bumping the major version.
    Example: 1 => 2.0.0, 1.2 => 2.0.0, 1.2.3 => 2.0.0

    # Examples from Poetry docs
    >>> get_caret_ceiling("0")
    '1.0.0'
    >>> get_caret_ceiling("0.0")
    '0.1.0'
    >>> get_caret_ceiling("0.0.3")
    '0.0.4'
    >>> get_caret_ceiling("0.2.3")
    '0.3.0'
    >>> get_caret_ceiling("1")
    '2.0.0'
    >>> get_caret_ceiling("1.2")
    '2.0.0'
    >>> get_caret_ceiling("1.2.3")
    '2.0.0'
    >>> get_caret_ceiling("1.2.3.post1")
    '2.0.0'
    >>> get_caret_ceiling("2!1.2.3.post1")
    '2.0.0'
    """
    if not isinstance(version, Version):
        version = Version(version)
    # Determine the upper bound
    if version.major > 0 or len(version.release) == 1:
        ceiling = f"{version.major + 1}.0.0"
    elif version.minor > 0 or len(version.release) == 2:
        ceiling = f"0.{version.minor + 1}.0"
    else:
        ceiling = f"0.0.{version.micro + 1}"
    return ceiling


def get_tilde_ceiling(version: str | Version) -> str:
    """
    Accepts a Poetry tilde target and returns the exclusive version ceiling.

    # Examples from Poetry docs
    >>> get_tilde_ceiling("1")
    '2.0.0'
    >>> get_tilde_ceiling("1.2")
    '1.3.0'
    >>> get_tilde_ceiling("1.2.3")
    '1.3.0'
    """
    if not isinstance(version, Version):
        version = Version(version)
    # Determine the upper bound based on the specified components
    if len(version.release) in [2, 3]:  # Major, Minor, Micro, or Major, Minor
        tilde_ceiling = f"{version.major}.{version.minor + 1}.0"
    else:  # Major, Minor or Only Major
        tilde_ceiling = f"{version.major + 1}.0.0"
    return tilde_ceiling


def encode_poetry_version(poetry_specifier: str) -> str:
    """
    Encodes Poetry version specifier as a Conda version specifier.

    Example: ^1 => >=1.0.0,<2.0.0

    # should be unchanged
    >>> encode_poetry_version("~=1.1")
    '~=1.1'
    >>> encode_poetry_version("1.*")
    '1.*'
    >>> encode_poetry_version(">=1,<2")
    '>=1,<2'
    >>> encode_poetry_version("==1.2.3")
    '==1.2.3'
    >>> encode_poetry_version("!=1.2.3")
    '!=1.2.3'
    >>> encode_poetry_version("===1.2.3")
    '===1.2.3'

    # strip spaces
    >>> encode_poetry_version(">= 1, < 2")
    '>=1,<2'

    # handle exact version specifiers correctly
    >>> encode_poetry_version("1.2.3")
    '1.2.3'
    >>> encode_poetry_version("==1.2.3")
    '==1.2.3'

    # handle caret operator correctly
    # examples from Poetry docs
    >>> encode_poetry_version("^0")
    '>=0.0.0,<1.0.0'
    >>> encode_poetry_version("^0.0")
    '>=0.0.0,<0.1.0'
    >>> encode_poetry_version("^0.0.3")
    '>=0.0.3,<0.0.4'
    >>> encode_poetry_version("^0.2.3")
    '>=0.2.3,<0.3.0'
    >>> encode_poetry_version("^1")
    '>=1.0.0,<2.0.0'
    >>> encode_poetry_version("^1.2")
    '>=1.2.0,<2.0.0'
    >>> encode_poetry_version("^1.2.3")
    '>=1.2.3,<2.0.0'

    # handle caret operator with a PEP440 version
    # correctly
    >>> encode_poetry_version("^0.8.post1")
    '>=0.8.0.post1,<0.9.0'

    # handle tilde operator correctly
    # examples from Poetry docs
    >>> encode_poetry_version("~1")
    '>=1.0.0,<2.0.0'
    >>> encode_poetry_version("~1.2")
    '>=1.2.0,<1.3.0'
    >>> encode_poetry_version("~1.2.3")
    '>=1.2.3,<1.3.0'

    # # handle tilde operator with a PEP440 version
    # # correctly
    # >>> encode_poetry_version("~0.8.post1")
    # '>=0.8.0.post1,<0.9.0.a0'

    # handle or operator correctly
    >>> encode_poetry_version("1.2.3|1.2.4")
    '1.2.3|1.2.4'
    >>> encode_poetry_version("^5|| ^6 | ^7")
    '>=5.0.0,<6.0.0|>=6.0.0,<7.0.0|>=7.0.0,<8.0.0'
    """
    if "|" in poetry_specifier:
        poetry_or_clauses = [clause.strip() for clause in poetry_specifier.split("|")]
        conda_or_clauses = [
            encode_poetry_version(clause)
            for clause in poetry_or_clauses
            if clause != ""
        ]
        return "|".join(conda_or_clauses)

    poetry_clauses = poetry_specifier.split(",")

    conda_clauses = []
    for poetry_clause in poetry_clauses:
        poetry_clause = poetry_clause.replace(" ", "")
        if poetry_clause.startswith("^"):
            # handle ^ operator
            caret_version = Version(poetry_clause[1:])
            floor = get_padded_base_version(caret_version)
            ceiling = get_caret_ceiling(caret_version)
            conda_clauses.append(">=" + floor)
            conda_clauses.append("<" + ceiling)
            continue

        if poetry_clause.startswith("~="):
            # handle the compatible release operator ~=
            # before the tilde ~ operator
            conda_clauses.append(poetry_clause)
            continue

        if poetry_clause.startswith("~"):
            # handle ~ operator
            tilde_version = poetry_clause[1:]
            floor = get_padded_base_version(tilde_version)
            ceiling = get_tilde_ceiling(tilde_version)
            conda_clauses.append(">=" + floor)
            conda_clauses.append("<" + ceiling)
            continue

        # other poetry clauses should be conda-compatible
        conda_clauses.append(poetry_clause)

    return ",".join(conda_clauses)


def encode_poetry_platform_to_selector_item(poetry_platform: str) -> str:
    """
    Encodes Poetry Platform specifier as a Conda selector.

    Example: "darwin" => "osx"
    """

    platform_selectors = {"windows": "win", "linux": "linux", "darwin": "osx"}
    poetry_platform = poetry_platform.lower().strip()
    if poetry_platform in platform_selectors:
        return platform_selectors[poetry_platform]
    else:  # unknown
        return ""


def encode_poetry_python_version_to_selector_item(poetry_specifier: str) -> str:
    """
    Encodes Poetry Python version specifier set as a Conda selector.

    Example:
        ">=3.8,<3.12" => "py>=38 and py<312"
        ">=3.8,<3.12,!=3.11" => "py>=38 and py<312 and py!=311"
        "<3.8|>=3.10" => "py<38 or py>=310"
        "<3.8|>=3.10,!=3.11" => "py<38 or py>=310 and py!=311"

    # handle exact version specifiers correctly
    >>> encode_poetry_python_version_to_selector_item("3")
    'py==3'
    >>> encode_poetry_python_version_to_selector_item("3.8")
    'py==38'
    >>> encode_poetry_python_version_to_selector_item("==3.8")
    'py==38'
    >>> encode_poetry_python_version_to_selector_item("!=3.8")
    'py!=38'
    >>> encode_poetry_python_version_to_selector_item("!=3.8.1")
    'py!=38'

    # handle caret operator correctly
    >>> encode_poetry_python_version_to_selector_item("^3.10") # '>=3.10.0,<4.0.0'
    'py>=310 and py<4'

    # handle tilde operator correctly
    >>> encode_poetry_python_version_to_selector_item("~3.10") # '>=3.10.0,<3.11.0'
    'py>=310 and py<311'

    # handle multiple requirements correctly (in "and")
    >>> encode_poetry_python_version_to_selector_item(">=3.8,<3.12,!=3.11")
    'py>=38 and py<312 and py!=311'

    # handle multiple requirements in "or" correctly ("and" takes precendence)
    >>> encode_poetry_python_version_to_selector_item("<3.8|>=3.10,!=3.11")
    'py<38 or py>=310 and py!=311'

    # handle compatible release operator correctly
    >>> encode_poetry_python_version_to_selector_item("~=3")
    'py>=3'
    >>> encode_poetry_python_version_to_selector_item("~=3.8")
    'py>=38 and py<4'
    >>> encode_poetry_python_version_to_selector_item("~=3.8.1")
    'py==38'
    >>> encode_poetry_python_version_to_selector_item("~=3.8.0.1")
    'py==38'
    >>> encode_poetry_python_version_to_selector_item("~=3.8,!=3.11")
    'py>=38 and py<4 and py!=311'

    # handle wildcard versions correctly
    >>> encode_poetry_python_version_to_selector_item("*")
    ''
    >>> encode_poetry_python_version_to_selector_item("3.*,!=3.11")
    'py>=3 and py<4 and py!=311'
    >>> encode_poetry_python_version_to_selector_item("!=3.*|3.11")
    'py<3 or py>=4 or py==311'
    >>> encode_poetry_python_version_to_selector_item("!=3.*,!=4.1")
    '(py<3 or py>=4) and py!=41'

    """

    if not poetry_specifier:
        return ""

    version_specifier = encode_poetry_version(poetry_specifier)

    if "|" in version_specifier:
        poetry_or_clauses = [clause.strip() for clause in version_specifier.split("|")]
        conda_or_clauses = [
            encode_poetry_python_version_to_selector_item(clause)
            for clause in poetry_or_clauses
            if clause != ""
        ]
        conda_or_clauses = " or ".join(conda_or_clauses)
        return conda_or_clauses

    conda_clauses = version_specifier.split(",")

    conda_selectors = []
    for conda_clause in conda_clauses:
        conda_selector = parse_python_version_specifier_to_selector(conda_clause)
        if conda_selector != "":
            conda_selectors.append(conda_selector)
    if len(conda_selectors) > 1:
        conda_selectors = [
            conda_selector if " or " not in conda_selector else f"({conda_selector})"
            for conda_selector in conda_selectors
        ]
    selectors = " and ".join(conda_selectors)
    return selectors


def parse_python_version_specifier_to_selector(version_specifier: str):
    """
    Take a Python version specifier, PEP 440 compliant.

    Return Python version conda selector.

    If the version_specifier has no operator, the equal operator ==
    is assumed.

    The version is normalized to "major.minor" (drop patch if present)
    or only "major" if minor is 0 (e.g. "3.8" -> "38", "3.8.1" -> "38",
    "3.0" -> "3", "3.0.1" -> "3").

    The compatible release operator ~= is expanded eventually into two selector
    items if the version has major and minor (e.g. "~=3.8" -> "py>=38 and py<4",
    but "~=3.8.1" -> "py==38").

    The exact equality operators == and != support the wildcard *
    in the version (e.g. "*" -> "==*" -> "").

    Examples:
        ">=3.8"   ->   "py>=38"
        "3.12"    ->   "py==312"
        "~=3.8"   ->   "py>=38 and py<4"
        "~=3.8.1" ->   "py==38"
        "3.*"     ->   "py>=3 and py<4"
        "!=3.*"   ->   "py<3 or py>=4"

    >>> parse_python_version_specifier_to_selector(">=3.8")
    'py>=38'
    >>> parse_python_version_specifier_to_selector("3.12")
    'py==312'
    >>> parse_python_version_specifier_to_selector("<4.0.0")
    'py<4'
    >>> parse_python_version_specifier_to_selector("<4.0.0.1")
    'py<4'
    >>> parse_python_version_specifier_to_selector(">=3")
    'py>=3'
    >>> parse_python_version_specifier_to_selector(">=3.8.0")
    'py>=38'
    >>> parse_python_version_specifier_to_selector(">=3.8.0.1")
    'py>=38'
    >>> parse_python_version_specifier_to_selector("~=3.8")
    'py>=38 and py<4'
    >>> parse_python_version_specifier_to_selector("3.*")
    'py>=3 and py<4'
    >>> parse_python_version_specifier_to_selector("!=3.*")
    'py<3 or py>=4'

    """
    # Regex to split an optional operator and a whatever version
    pattern = r"^(?P<operator>\^|~=|~|>=|<=|>|<|!=|===|==|=)?(?P<version>.+)$"

    # Here Specifier or Version are not useful because
    # Specifier requires an operator, and Version cannot
    # accept an operator. Doomed to match twice.

    match = re.match(pattern, version_specifier)
    if not match:
        raise ValueError(f"Invalid version selector: {version_specifier}")

    # Extract operator and version
    operator = match.group("operator")
    version = match.group("version")

    if operator in [None, "=", "==", "==="]:
        # Default to "==" if no operator is provided or "=", "==="
        operator = "=="
        # Check also if there is a wildcard operator "*" (may result in two operators)
        return expand_operator_wildcard_version_to_selector(operator, version)
    elif operator == "~=":
        # Compatible release operator "~=" (may result in two operators)
        return expand_compatible_release_operator_version_to_selector(version)
    elif operator == "!=":
        # Check also if there is a wildcard operator "*" (may result in two operators)
        return expand_operator_wildcard_version_to_selector(operator, version)
    return operator_version_to_selector(operator, Version(version))


def expand_compatible_release_operator_version_to_selector(
    version: str | Version,
) -> str:
    """
    Take a Python version, PEP440 compliant.

    The compatible release operator "~=" is implicit.

    The python version should be reasonable and realistic (e.g. "3.11"), but
    it is true that an esoteric still PEP440 valid version would make grayskull
    crash, therefore here we parse the Python version as a PEP440 compliant
    version (e.g. "3.11.3.dev0").

    The compatible release operator ~= is expanded eventually into two selector
    items if the version has major and minor (e.g. "~=3.8" -> "py>=38 and py<4",
    but "~=3.8.1" -> "py==38", and "~=3" -> "py>=38").

    The reason why this operator is expanded here and not in
    encode_poetry_python_version_to_selector_item just after
    encode_poetry_version is because the selectors for the python
    version use only major and minor, and therefore the compatible
    release operator ~= makes sense only in the case of major and
    minor specified (e.g. "~=3.8" -> "py>=38 and py<4") and just by
    knowing that we can avoid having to detect cases like:
        "~=3.8.0.1" -> ">=3.8.0.1, ==3.8.0.*" -> ">=3.8.0.1, <3.8.1.0a" -> "py==38"
    and expand only for:
        "~=3.8" -> ">=3.8, ==3.*" -> ">=3.8, <4.0a" -> "py>=38 and py<4"
        "~=3.0" -> ">=3.0, ==3.*" -> ">=3.0, <4.0a" -> "py>=3 and py<4"
    in the rest of the cases it's a simple conversion to ">=" operator.

    If we would expand it before, we would receive specifier sets like
    ">=3.8.0.1, <3.8.1.0a" (among other specifiers) and we would need to
    detect those cases to avoid rendering to a naive "py>=38 and py<38"
    which would be an invalid statement.

    Rationale:
        - generally it would work in this way:
            ~=2     -> illegal for PEP440
            ~=2.2   -> ">=2.2, ==2.*" -> ">=2.2, <3.0a"
            ~=1.4.5 -> ">=1.4.5, ==1.4.*" -> ">=1.4.5, <1.5.0a"
            ~=0.5.3 -> ">=0.5.3, ==0.5.*" -> ">=0.5.3, <0.6.0a"
        - considering only python versions and their selectors:
            ~=3       -> illegal for PEP440     -> ">=3, ==*" -> ">=3" -> "py>=3"
            ~=3.8     -> ">=3.8, ==3.*"         -> ">=3.8, <4.0a" -> "py>=38 and py<4"
            ~=3.8.1   -> ">=3.8.1, ==3.8.*"     -> ">=3.8.1, <3.9.0a" -> "py==38"
            ~=3.8.0.1 -> ">=3.8.0.1, ==3.8.0.*" -> ">=3.8.0.1, <3.8.1.0a" -> "py==38"

    Examples:
        "3"     ->   "py>=3"
        "3.8"   ->   "py>=38 and py<4"
        "3.8.1" ->   "py==38"

    >>> expand_compatible_release_operator_version_to_selector("3")
    'py>=3'
    >>> expand_compatible_release_operator_version_to_selector("3.8")
    'py>=38 and py<4'
    >>> expand_compatible_release_operator_version_to_selector("3.0")
    'py>=3 and py<4'
    >>> expand_compatible_release_operator_version_to_selector("3.8.1")
    'py==38'
    >>> expand_compatible_release_operator_version_to_selector("3.8.1.1")
    'py==38'
    >>> expand_compatible_release_operator_version_to_selector("3.8a0")
    'py>=38 and py<4'
    >>> expand_compatible_release_operator_version_to_selector("3.8.1.post1")
    'py==38'
    """
    if not isinstance(version, Version):
        version = Version(version)

    # The compatible release operator ~= is expanded eventually
    # into two selector items if the version has major and minor
    # even if the minor is 0, because it would be used as a padding
    # placeholder.
    # See:
    # https://packaging.python.org/en/latest/specifications/version-specifiers/#compatible-release
    if len(version.release) < 3:
        # "3.8" -> "py>=38 and py<4", and "3" -> "py>=3"
        lower_bound_operator = ">="
    else:
        # "3.8.1" -> "py==38"
        lower_bound_operator = "=="
    lower_bound_selector = operator_version_to_selector(lower_bound_operator, version)
    if len(version.release) == 2:
        # get version selector, with ">=" operator as lower bound
        # get the ceiling of the version (major bumped by 1)
        ceiling_version = version.major + 1
        # get ceiling version selector, with "<" operator as upper bound
        upper_bound = operator_version_to_selector("<", Version(f"{ceiling_version}"))
        return f"{lower_bound_selector} and {upper_bound}"
    return lower_bound_selector


def expand_operator_wildcard_version_to_selector(
    operator: str | None, version: str | Version
) -> str:
    """
    Take the strict equality operators "==" or "!=" and
    a Python version ending with ".*", PEP440 compliant.

    "*" is accepted, but "1*" or "1.1*" are not accepted
    because PEP 440 requires the "*" wildcard to follow a "."
    because it is meant to represent a "range of versions
    with common prefix components."

    Wildcards can be expressed as ranges (">=" and "<") using
    the next significant component.

    Examples:
         "*" -> "==*" -> ""
         "1.*" -> ">=1.0.0.a0,<2.0.0"
         "1.1.*" -> ">=1.1.0.a0,<1.2.0"
         "1.1.1.*" -> ">=1.1.1.a0,<1.1.2"

    inclusion: 1.1
            == 1.1        # Equal, so 1.1 matches clause
            == 1.1.0      # Zero padding expands 1.1 to 1.1.0, so it matches clause
            == 1.1.dev1   # Not equal (dev-release), so 1.1 does not match clause
            == 1.1a1      # Not equal (pre-release), so 1.1 does not match clause
            == 1.1.post1  # Not equal (post-release), so 1.1 does not match clause
            == 1.1.*      # Same prefix, so 1.1 matches clause

    exclusion: 1.1.post1
            != 1.1        # Not equal, so 1.1.post1 matches clause
            != 1.1.post1  # Equal, so 1.1.post1 does not match clause
            != 1.1.*      # Same prefix, so 1.1.post1 does not match clause

    # In practice, if the star suffix ".*" is used on Python version specifiers
    # to be rendered as conda selector, we can simplify the calculation according
    # to which operator is used:
    #
    # - equality: it makes sense to expand it only if the version is "{major}.*"
    #             (e.g. "3.*" -> ">=3.0a0,<4" -> "py>=3 and py<4"). In all the
    #             other cases it is enough to remove the ".*" and consider as
    #             usual the "{operator}{major}{minor}" if "minor" is more than
    #             "0", otherwise "{operator}{major}".

    # - inequality: it makes sense to expand it only if the version is "{major}.*"
    #             (e.g. "3.*" -> ">=3.0a0,<4" -> "py>=3 and py<4"). In all the
    #             other cases it is enough to remove the ".*" and consider as
    #             usual the "{operator}{major}{minor}" if "minor" is more than
    #             "0", otherwise "{operator}{major}".

    # Equality examples

    >>> expand_operator_wildcard_version_to_selector("==","*") # any
    ''

    >>> expand_operator_wildcard_version_to_selector("==", "3.*") # >=3.0a0,<4
    'py>=3 and py<4'

    # >=3.12.0a0,<3.13
    >>> expand_operator_wildcard_version_to_selector("==", "3.12.*")
    'py==312'

    # >=3.9.1.0a0,<3.9.2
    >>> expand_operator_wildcard_version_to_selector("==", "3.9.1.*")
    'py==39'

    # >=3.9.1.0a0,<3.9.1.2
    >>> expand_operator_wildcard_version_to_selector("==", "3.9.1.1.*")
    'py==39'

    # Inequality examples

    >>> expand_operator_wildcard_version_to_selector("!=","*") # none
    'py<0'

    # <3.0a0|>=4
    >>> expand_operator_wildcard_version_to_selector("!=", "3.*")
    'py<3 or py>=4'

    # <3.12.0a0|>=3.13
    >>> expand_operator_wildcard_version_to_selector("!=", "3.12.*")
    'py<312 or py>=313'

    # <3.9.1.0a0|>=3.9.2
    >>> expand_operator_wildcard_version_to_selector("!=", "3.9.1.*")
    'py<39 or py>=39'

    # <3.9.1.1.0a0|>=3.9.1.2
    >>> expand_operator_wildcard_version_to_selector("!=", "3.9.1.1.*")
    'py<39 or py>=39'

    """
    if version == "*":
        # This should not happen, as the "*" is stripped away before
        # to avoid having trivial selectors, but consider it anyway
        # for general usage.
        if operator in [None, "", "=", "==", "==="]:
            return ""
        else:
            return "py<0"
    base_version = version.rstrip(".*")
    expand_operator_wildcard_version = len(base_version) != len(version)
    version = Version(base_version)
    if operator in [None, "", "=", "==", "==="]:
        # Default to "==" if no operator is provided or "=", "==="
        operator = "=="
        # it makes sense to expand it only if the version is "{major}.*"
        if expand_operator_wildcard_version and len(version.release) == 1:
            # "3.*" -> ">=3.0a0,<4" -> "py>=3 and py<4"
            left_bound_selector = operator_version_to_selector(">=", version)
            # get the ceiling of the version (major bumped by 1)
            right_version = version.major + 1
            # get ceiling version selector, with "<" operator as upper bound
            right_bound_selector = operator_version_to_selector(
                "<", Version(f"{right_version}")
            )
            return f"{left_bound_selector} and {right_bound_selector}"
    elif operator == "!=":
        if expand_operator_wildcard_version:
            left_bound_selector = operator_version_to_selector("<", version)
            if len(version.release) == 1:
                # "3.*" -> "<3.0a0|>=4" -> "py<3 or py>=4"
                # major bumped by 1
                right_version = version.major + 1
            elif len(version.release) == 2:
                # "3.12.*" -> "<3.12.0a0|>=3.13" -> "py<312 or py>=313"
                # minor bumped by 1
                right_version = f"{version.major}." + str(version.minor + 1)
            else:
                # "3.9.1.*" -> <3.9.1.0a0|>=3.9.2"" -> "py<39 or py>=39"
                # "3.9.1.1.*" -> <3.9.1.1.0a0|>=3.9.1.2"" -> "py<39 or py>=39"
                # use the same version in the right bound
                right_version = f"{version.major}.{version.minor}"
            # get ceiling version selector, with ">=" operator as upper bound
            right_bound_selector = operator_version_to_selector(
                ">=", Version(f"{right_version}")
            )
            return f"{left_bound_selector} or {right_bound_selector}"
    return operator_version_to_selector(operator, version)


def operator_version_to_selector(operator: str | None, version: str | Version) -> str:
    """
    Consider major, minor, and discard the rest (patch or additional parts)
    Return only major if minor is "0", otherwise return major.minor

    >>> operator_version_to_selector(">=", "3.8")
    'py>=38'
    >>> operator_version_to_selector("==", "3.12")
    'py==312'
    >>> operator_version_to_selector("", "3.12")
    'py==312'
    >>> operator_version_to_selector("<", "4.0.0")
    'py<4'
    >>> operator_version_to_selector("<", "4.0.0.1")
    'py<4'
    >>> operator_version_to_selector(">=", "3")
    'py>=3'
    >>> operator_version_to_selector(">=", "3.8.0")
    'py>=38'
    >>> operator_version_to_selector(">=", "3.8.0.1")
    'py>=38'
    >>> operator_version_to_selector(">=", "3.8.0.1post1")
    'py>=38'
    >>> operator_version_to_selector(">=", "3.8.0.1a0")
    'py>=38'
    >>> operator_version_to_selector("<", "2!4.0.0.1.post1")
    'py<4'
    """
    if operator in [None, "", "=", "==="]:
        # Default to "==" if no operator is provided or "=", "==="
        operator = "=="
    if not isinstance(version, Version):
        version = Version(version)
    version_selector = (
        version.major if version.minor == 0 else f"{version.major}{version.minor}"
    )
    return f"py{operator}{version_selector}"


def combine_conda_selectors(python_selector: str, platform_selector: str):
    """
    Combine selectors based on presence
    """
    if python_selector and platform_selector:
        if " or " in python_selector:
            python_selector = f"({python_selector})"
        selector = f"{python_selector} and {platform_selector}"
    elif python_selector:
        selector = f"{python_selector}"
    elif platform_selector:
        selector = f"{platform_selector}"
    else:
        selector = ""
    return f"  # [{selector}]" if selector else ""



from __future__ import annotations

import re
from abc import ABC
from pathlib import Path

from souschef.ingredient import Ingredient
from souschef.jinja_expression import get_global_jinja_var
from souschef.recipe import Recipe

from grayskull.strategy.cran import CranStrategy
from grayskull.strategy.pypi import PypiStrategy


class GrayskullFactory(ABC):
    REGISTERED_STRATEGY = {
        "pypi": PypiStrategy,
        "cran": CranStrategy,
    }

    @staticmethod
    def create_recipe(repo_type: str, config, pkg_name=None, sections_populate=None):
        if repo_type.lower() not in GrayskullFactory.REGISTERED_STRATEGY:
            raise ValueError(
                f"Recipe generator {repo_type.lower()} does not exist.\n"
                f"Please inform a valid one."
                f"{', '.join(GrayskullFactory.REGISTERED_STRATEGY.keys())}"
            )
        if Path(config.name).is_file() and not config.from_local_sdist:
            recipe = Recipe(load_file=config.name)
            recipe_name = _get_name(recipe)
            recipe_version = _get_version(recipe)
            config.name = (
                recipe_name.value
                if isinstance(recipe_name, Ingredient)
                else recipe_name
            )
            config.version = (
                recipe_version.value
                if isinstance(recipe_version, Ingredient)
                else recipe_version
            )
        else:
            pkg_name = pkg_name or config.name
            recipe = Recipe(name=pkg_name, version=config.version)
        if config.name.startswith(("<{", "{{", "r-{{", "r-<{")):
            config.name = get_global_jinja_var(recipe, "name")
        GrayskullFactory.REGISTERED_STRATEGY[repo_type.lower()].fetch_data(
            recipe, config, sections=sections_populate
        )

        if "build" not in recipe:
            recipe.add_section({"build": {"number": 0}})
        else:
            recipe["build"]["number"] = 0

        return recipe


def _get_name(recipe):
    return __get_var(recipe, "name")


def __get_var(recipe, val):
    if recipe["package"][val].value.strip().startswith("<{"):
        re_jinja_var = re.match(r"\s*<{\s*(\w+)", recipe["package"][val].value)
        if re_jinja_var:
            jinja_var = re_jinja_var.groups()[0]
            try:
                return get_global_jinja_var(recipe, jinja_var)
            except ValueError:
                return None
    return recipe["package"][val]


def _get_version(recipe):
    return __get_var(recipe, "version")



import re
from pathlib import Path

from grayskull.utils import origin_is_github, origin_is_local_sdist


def parse_pkg_name_version(
    pkg_name: str,
) -> tuple[str, str, str | None]:
    origin = ""
    if origin_is_local_sdist(pkg_name):
        # Try to get package name and version from sdist archive
        # If the version is normalized, there should be no dash in it
        # Will get them from PKG-INFO later
        filename = Path(pkg_name).stem
        if filename.endswith(".tar"):
            filename = filename[:-4]
        name, _, version = filename.rpartition("-")
        if name == "":
            name = filename
            version = ""
        return "", name, version
    if origin_is_github(pkg_name):
        origin, pkg_name = pkg_name.rsplit("/", 1)
        origin += "/"
        if pkg_name.endswith(".git"):
            pkg_name = pkg_name[:-4]
    pkg = re.match(r"([a-zA-Z0-9\-_\.]+)=+([a-zA-Z0-9\-_\.]+)", pkg_name)
    if pkg:
        pkg_name = origin + pkg.group(1)
        version = pkg.group(2)
        return "", pkg_name, version
    return origin, pkg_name, None



import os
from functools import lru_cache


@lru_cache(maxsize=2)
def get_all_licenses() -> list:
    data_folder = os.path.dirname(__file__)
    all_licenses = []
    for license_file in os.listdir(data_folder):
        full_path = os.path.join(data_folder, license_file)
        if not os.path.isfile(full_path) or license_file.endswith(".py"):
            continue
        with open(full_path) as f:
            all_licenses.append((license_file, f.read()))
    return all_licenses



from __future__ import annotations

import base64
import json
import logging
import os
import re
from copy import deepcopy
from dataclasses import dataclass
from functools import lru_cache
from operator import itemgetter
from pathlib import Path
from subprocess import check_output
from tempfile import mkdtemp

import requests
from colorama import Fore
from rapidfuzz import process
from rapidfuzz.distance import OSA
from rapidfuzz.fuzz import partial_ratio, token_set_ratio, token_sort_ratio

from grayskull.cli.stdout import print_msg
from grayskull.license.data import get_all_licenses  # noqa

log = logging.getLogger(__name__)


@dataclass
class ShortLicense:
    name: str
    path: str | Path | None
    is_packaged: bool


@lru_cache(maxsize=10)
def get_all_licenses_from_spdx() -> list:
    """Get all licenses available on spdx.org

    :return: List with all licenses information on spdx.org
    """
    try:
        response = requests.get(
            url="https://spdx.org/licenses/licenses.json", timeout=5
        )
    except requests.exceptions.ConnectionError:
        log.info(
            "SPDX licence server didn't respond. Grayskull will continue without that."
        )
        return []
    log.debug(
        f"Response from spdx.org. Status code:{response.status_code},"
        f" response: {response}"
    )
    if response.status_code != 200:
        raise requests.HTTPError(
            f"It was not possible to communicate with spdx api.\n{response.text}"
        )
    print_msg("Recovering license info from spdx.org ...")
    return [
        lic
        for lic in response.json()["licenses"]
        if not lic.get("isDeprecatedLicenseId", False)
    ]


def _match_scrambled_exact(candidate, licenses) -> str | None:
    """
    Return license with rearranged word order only.

    Fancy scorer confuses BSD-3-Clause with DEC-3-Clause.
    """
    bag = set(re.findall(r"\w+", candidate.lower()))
    for license in licenses:
        if bag == set(re.findall(r"\w+", license.lower())):
            return license
    return None


def match_license(name: str) -> dict:
    """Match if the given license name matches any license present on
    spdx.org

    :param name: License name
    :return: Information of the license matched
    """
    all_licenses = get_all_licenses_from_spdx()
    if not all_licenses:
        return {}
    name = re.sub(r"\s+license\s*", "", name.strip(), flags=re.IGNORECASE)
    name = name.strip()

    exact_match = _match_scrambled_exact(name, _get_all_license_choice(all_licenses))
    if exact_match:
        best_matches = [(exact_match, 100, 0)]
        spdx_license = best_matches[0]
    else:
        best_matches = process.extract(
            name, _get_all_license_choice(all_licenses), scorer=partial_ratio
        )
        best_matches = process.extract(name, [lc for lc, *_ in best_matches])
        spdx_license = best_matches[0]

    if spdx_license[1] < 100:
        # Prefer "-or-later" licenses over the "-only"
        later_licenses = {
            lic[0].replace("-or-later", "")
            for lic in best_matches
            if lic[0].endswith("-or-later")
        }
        best_matches = [
            lic[0]
            for lic in best_matches
            if not (
                lic[0].endswith("-only")
                and lic[0].replace("-only", "") in later_licenses
            )
        ]

        if best_matches:
            best_matches = process.extract(
                name, best_matches, scorer=OSA.normalized_similarity
            )
            original_matches = deepcopy(best_matches)

            if name.startswith("GPL"):
                original_matches = [
                    m for m in original_matches if m[0].startswith("GPL")
                ]
            spdx_license = best_matches[0]
            best_matches = [
                lic[0] for lic in original_matches if lic[1] >= spdx_license[1]
            ]
            if len(best_matches) > 1:
                spdx_license = process.extractOne(
                    name, best_matches, scorer=token_sort_ratio
                )
            if original_matches and original_matches[0][1] < 0.55:
                spdx_license = process.extractOne(
                    name, [m[0] for m in original_matches], scorer=token_sort_ratio
                )

    if spdx_license[1] != 100 and spdx_license[0].startswith("MIT"):
        spdx_license = "MIT"
    else:
        spdx_license = spdx_license[0]

    log.info(
        f"Best match for license {name} was {spdx_license}.\n"
        f"Best matches: {best_matches}"
    )

    return _get_license(spdx_license, all_licenses)


def get_short_license_id(name: str) -> str:
    """Get the spdx identifier for the given license name

    :param name: License name
    :return: short identifier (spdx) for the given license name
    """
    recipe_license = match_license(name)
    if not recipe_license:
        return "IT-WAS-NOT-POSSIBLE-TO-RECOVER-LICENCE"
    return recipe_license["licenseId"]


def _get_license(license_id: str, all_licenses: list) -> dict:
    """Search for the license identification in all licenses received

    :param license_id: license identification
    :param all_licenses: List with all licenses
    :return: Dict with the information of the license desired
    """
    for one_license in all_licenses:
        if license_id in _get_all_names_from_api(one_license):
            return one_license


def _get_all_names_from_api(one_license: dict) -> list:
    """Get the names and other names which each license has.

    :param one_license: License name
    :return: List of all names which the given license is know of
    """
    result = set()
    if one_license["name"]:
        result.add(one_license["name"])
    if one_license["licenseId"]:
        result.add(one_license["licenseId"])
    other_names = get_other_names_from_opensource(one_license["licenseId"])
    result.update(other_names)
    return list(result)


def get_other_names_from_opensource(license_spdx: str) -> list:
    lic = get_opensource_license(license_spdx)
    return [_license["name"] for _license in lic.get("other_names", [])]


def get_opensource_license(license_spdx: str) -> dict:
    opensource = get_opensource_license_data()
    for lic in opensource:
        if lic["id"] == license_spdx:
            return lic
        for _id in lic["identifiers"]:
            if _id["scheme"].lower() == "spdx" and license_spdx == _id["identifier"]:
                return lic
    return {}


def read_licence_cache():
    with open(Path(__file__).parent / "licence_cache.json") as licence_cache:
        return json.load(licence_cache)


@lru_cache(maxsize=10)
def get_opensource_license_data() -> list:
    try:
        response = requests.get(url="https://api.opensource.org/licenses/", timeout=5)
    except requests.exceptions.RequestException:
        return read_licence_cache()
    if response.status_code != 200:
        return read_licence_cache()
    return response.json()


def _get_all_license_choice(all_licenses: list) -> list:
    """Function responsible to get the whole licenses name

    :param all_licenses: list with all licenses
    :return: list with all names which each license may have
    """
    all_choices = []
    for api_license in all_licenses:
        all_choices += _get_all_names_from_api(api_license)
    return all_choices


def search_license_file(
    folder_path: str,
    git_url: str | None = None,
    version: str | None = None,
    license_name_metadata: str | None = None,
    folders_exclude_search: tuple[str, ...] = tuple(),
) -> list[ShortLicense]:
    """Search for the license file. First it will try to find it in the given
    folder, after that it will search on the github api and for the last it will
    clone the repository and it will search for the license there.

    :param folder_path: Path where the sdist package was unpacked
    :param git_url: URL for the Github repository
    :param version: Package version
    :param license_name_metadata: Default value for the license type
    :return: ShortLicense with the information regarding the license
    """
    if license_name_metadata:
        license_name_metadata = get_short_license_id(license_name_metadata)

    all_license_sdist = search_license_folder(
        folder_path, license_name_metadata, folders_exclude_search
    )
    for license_sdist in all_license_sdist:
        license_sdist.is_packaged = True
        license_sdist.path = os.path.relpath(license_sdist.path, folder_path)
        license_sdist.path = license_sdist.path.replace("\\", "/")

        splited = license_sdist.path.split("/")
        if len(splited) > 1:
            license_sdist.path = "/".join(splited[1:])
    if all_license_sdist:
        return all_license_sdist

    if not git_url:
        return [ShortLicense(license_name_metadata, None, False)]

    github_license = search_license_api_github(git_url, version, license_name_metadata)
    if github_license:
        return [github_license]

    repo_license = search_license_repo(
        git_url, version, license_name_metadata, folders_exclude_search
    )
    if repo_license:
        return repo_license
    return [ShortLicense(license_name_metadata, None, False)]


@lru_cache(maxsize=13)
def search_license_api_github(
    github_url: str, version: str | None = None, default: str | None = "Other"
) -> ShortLicense | None:
    """Search for the license asking in the github api

    :param github_url: GitHub URL
    :param version: Package version
    :param default: default license type
    :return: License information
    """
    if github_url.endswith("/"):
        github_url = github_url[:-1]
    github_url = _get_api_github_url(github_url, version)
    log.info(f"Github url: {github_url} - recovering license info")
    print_msg("Recovering license information from github...")

    response = requests.get(url=github_url, timeout=10)
    if response.status_code != 200:
        return None

    json_content = response.json()
    license_content = base64.b64decode(json_content["content"]).decode("utf-8")
    file_path = os.path.join(mkdtemp(prefix="github-license-"), json_content["name"])
    with open(file_path, "w") as f:
        f.write(license_content)
    return ShortLicense(
        json_content.get("license", {}).get("spdx_id", default), file_path, False
    )


def _get_api_github_url(github_url: str, version: str | None = None) -> str:
    """Try to presume the github url

    :param github_url: GitHub URL
    :param version: package version
    :return: GitHub URL
    """
    github_url = re.sub(r"github.com", "api.github.com/repos", github_url)
    if github_url[-1] != "/":
        github_url += "/"

    github_url += "license"
    return f"{github_url}?ref={version}" if version else github_url


def search_license_folder(
    path: str | Path,
    default: str | None = None,
    folders_exclude_search: tuple[str, ...] = tuple(),
) -> list[ShortLicense]:
    """Search for the license in the given folder

    :param path: Sdist folder
    :param default: Default value for the license type
    :return: License information
    """
    folders_exclude_search = set(
        list(folders_exclude_search) + ["doc", "theme", "themes", "docs"]
    )
    re_search = re.compile(
        r"(\bcopyright\b|\bnotice\b|\blicense[s]*\b|\bcopying\b|\bcopyleft\b)",
        re.IGNORECASE,
    )
    all_licences = []
    for folder_path, dirnames, filenames in os.walk(str(path)):
        dirnames[:] = [
            folder
            for folder in dirnames
            if folder not in folders_exclude_search and not folder.startswith(".")
        ]
        for one_file in filenames:
            if re_search.match(one_file):
                lc_path = os.path.join(folder_path, one_file)
                all_licences.append(
                    ShortLicense(get_license_type(lc_path, default), lc_path, False)
                )
    return all_licences


def search_license_repo(
    git_url: str,
    version: str | None,
    default: str | None = None,
    folders_exclude_search: tuple[str, ...] = tuple(),
) -> list[ShortLicense] | None:
    """Search for the license file in the given github repository

    :param git_url: GitHub URL
    :param version: Package version
    :param default: Default value for the license type
    :param folders_exclude_search: Folders names to be excluded from search for licences
    :return: License information
    """
    git_url = re.sub(r"/$", ".git", git_url)
    git_url = git_url if git_url.endswith(".git") else f"{git_url}.git"
    print_msg("Recovering license info from repository...")
    tmp_dir = mkdtemp(prefix="gs-clone-repo-")
    try:
        check_output(_get_git_cmd(git_url, version, tmp_dir))
    except Exception as err:
        log.debug(
            f"Exception occurred when gs was trying to clone the repository."
            f" url: {git_url}, version: {version}. Exception: {err}"
        )
        if not version.startswith("v"):
            return search_license_repo(
                git_url, f"v{version}", default, folders_exclude_search
            )
        return None
    return search_license_folder(
        str(tmp_dir), default, folders_exclude_search=folders_exclude_search
    )


def _get_git_cmd(git_url: str, version: str, dest) -> list[str]:
    """Return the full git command to clone the repository

    :param git_url: GitHub URL
    :param version: Package version
    :param dest: Folder destination
    :return: git command to clone the repository
    """
    git_cmd = ["git", "clone"]
    if version:
        git_cmd += ["-b", version]
    return git_cmd + [git_url, str(dest)]


def get_license_type(path_license: str, default: str | None = None) -> str | None:
    """Function tries to match the license with one of the models present in
    grayskull/license/data

    :param path_license: Path to the license file
    :param default: Default value for the license type
    :return: License type
    """
    with open(path_license, errors="ignore") as license_file:
        license_content = license_file.read()
    find_apache = re.findall(
        r"apache\.org\/licenses\/LICENSE\-([0-9])\.([0-9])",
        license_content,
        re.IGNORECASE,
    )
    if find_apache:
        lic_type = find_apache[0]
        return f"Apache-{lic_type[0]}.{lic_type[1]}"
    print_msg("Matching license file with database from Grayskull...")
    all_licenses = get_all_licenses()
    licenses_text = list(map(itemgetter(1), all_licenses))
    best_match = process.extract(
        license_content, licenses_text, scorer=token_sort_ratio
    )
    print_msg(
        f"{Fore.YELLOW}Match percentage of the license is {int(best_match[0][1])}%. "
        + "Low match percentage could mean that the license was modified."
    )

    if default and best_match[0][1] < 51:
        log.info(f"Match too low for recipe {best_match}, using the default {default}")
        return default

    higher_match = best_match[0]
    equal_values = [val[0] for val in best_match if val[1] > (higher_match[1] - 3)]
    if len(equal_values) > 1:
        higher_match = process.extractOne(
            license_content, equal_values, scorer=token_set_ratio
        )
    index_license = licenses_text.index(higher_match[0])
    return all_licenses[index_license][0]
